@@ -1,12 +1,14 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+from decimal import Decimal
 from functools import wraps
 
 from trytond.i18n import gettext
-from trytond.model import Workflow, ModelView, fields
+from trytond.model import ModelView, Workflow, fields
 from trytond.model.exceptions import AccessError
-from trytond.transaction import Transaction
+from trytond.modules.product import round_price
 from trytond.pool import Pool, PoolMeta
+from trytond.transaction import Transaction
 
 
 def process_sale(moves_field):
@@ -15,12 +17,16 @@ def process_sale(moves_field):
         def wrapper(cls, shipments):
             pool = Pool()
             Sale = pool.get('sale.sale')
-            with Transaction().set_context(_check_access=False):
+            transaction = Transaction()
+            context = transaction.context
+            with transaction.set_context(_check_access=False):
                 sales = set(m.sale for s in cls.browse(shipments)
                     for m in getattr(s, moves_field) if m.sale)
             func(cls, shipments)
             if sales:
-                Sale.__queue__.process(sales)
+                with transaction.set_context(
+                        queue_batch=context.get('queue_batch', True)):
+                    Sale.__queue__.process(sales)
         return wrapper
     return _process_sale
 
@@ -35,7 +41,7 @@ class ShipmentOut(metaclass=PoolMeta):
         SaleLine = Pool().get('sale.line')
         for shipment in shipments:
             for move in shipment.outgoing_moves:
-                if (move.state == 'cancel'
+                if (move.state == 'cancelled'
                         and isinstance(move.origin, SaleLine)):
                     raise AccessError(
                         gettext('sale.msg_sale_move_reset_draft',
@@ -52,7 +58,7 @@ class ShipmentOut(metaclass=PoolMeta):
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('cancel')
+    @Workflow.transition('cancelled')
     @process_sale('outgoing_moves')
     def cancel(cls, shipments):
         super(ShipmentOut, cls).cancel(shipments)
@@ -68,7 +74,7 @@ class ShipmentOutReturn(metaclass=PoolMeta):
         SaleLine = Pool().get('sale.line')
         for shipment in shipments:
             for move in shipment.incoming_moves:
-                if (move.state == 'cancel'
+                if (move.state == 'cancelled'
                         and isinstance(move.origin, SaleLine)):
                     raise AccessError(
                         gettext('sale.msg_sale_move_reset_draft',
@@ -89,11 +95,15 @@ def process_sale_move(func):
     def wrapper(cls, moves):
         pool = Pool()
         Sale = pool.get('sale.sale')
-        with Transaction().set_context(_check_access=False):
+        transaction = Transaction()
+        context = transaction.context
+        with transaction.set_context(_check_access=False):
             sales = set(m.sale for m in cls.browse(moves) if m.sale)
         func(cls, moves)
         if sales:
-            Sale.__queue__.process(sales)
+            with transaction.set_context(
+                    queue_batch=context.get('queue_batch', True)):
+                Sale.__queue__.process(sales)
     return wrapper
 
 
@@ -153,6 +163,33 @@ class Move(metaclass=PoolMeta):
             category = self.origin.unit.category.id
         return category
 
+    def get_cost_price(self, product_cost_price=None):
+        pool = Pool()
+        SaleLine = pool.get('sale.line')
+        Sale = pool.get('sale.sale')
+        # For return sale's move use the cost price of the original sale
+        if (isinstance(self.origin, SaleLine)
+                and self.origin.quantity < 0
+                and self.from_location.type != 'storage'
+                and self.to_location.type == 'storage'
+                and isinstance(self.origin.sale.origin, Sale)):
+            sale = self.origin.sale.origin
+            cost = Decimal(0)
+            qty = Decimal(0)
+            for move in sale.moves:
+                if (move.state == 'done'
+                        and move.from_location.type == 'storage'
+                        and move.to_location.type == 'customer'
+                        and move.product == self.product):
+                    move_quantity = Decimal(str(move.internal_quantity))
+                    cost_price = move.get_cost_price(
+                        product_cost_price=move.cost_price)
+                    qty += move_quantity
+                    cost += cost_price * move_quantity
+            if qty:
+                product_cost_price = round_price(cost / qty)
+        return super().get_cost_price(product_cost_price=product_cost_price)
+
     @property
     def origin_name(self):
         pool = Pool()
@@ -164,7 +201,7 @@ class Move(metaclass=PoolMeta):
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('cancel')
+    @Workflow.transition('cancelled')
     @process_sale_move
     def cancel(cls, moves):
         super(Move, cls).cancel(moves)
